
;; ------------------------------------------------------------
;; William R. Bevier and Michael K. Smith.
;; Copyright 1994, Computational Logic, Inc. All Rights Reserved
;; ------------------------------------------------------------

(require "iterate")
(provide "general")

(defparameter *report-mode* t)

(eval-when
 (compile load eval)

 (defvar *base-directory* #.(namestring (probe-file "./")))

 (defun load-base (s)
   (let ((dir (pathname-directory s))
	 (type (pathname-type s)))
     (if (null dir)
	 (setq s (concatenate 'string *base-directory* s)))
     (if type
	 (load s)
	 (let ((obj (concatenate 'string s ".o"))
	       (src (concatenate 'string s ".lisp")))
	   (cond ((and (probe-file src)
		       (probe-file obj))
		  (if (< (file-write-date src) (file-write-date obj))
		      (load obj)
		      (load src)))
		 ((probe-file obj) (load obj))
		 ((probe-file src) (load src)))))))

 (defmacro f-l (args body)
   `(function (lambda ,args ,body)))

 (defmacro mv-let (args source body)
   `(multiple-value-bind ,args ,source ,body))

 (defmacro implies (a b)
   `(if ,a ,b t))

 (defvar *static-break* nil)
 (defmacro static-check (form string &rest args)
   `(if (not ,form)
	(if *static-break*
	    (error ,string ,@args)
	    (progn (status-report ,string ,@args) t))
	t))

 (defvar *eval-mode* nil)
 (defmacro eval-wrapper (form &key debug step debug-fn step-fn)
   `(let ((result ,form))		
      (case *eval-mode*
	(interactive (if ,step
			 (apply (or ,step-fn #'defaut-wrapper-print)
				(list ,step result))))
	(robot (if ,step
			 (apply (or ,step-fn #'defaut-wrapper-print)
				(list ,step result))))
	(debug (if ,debug
		   (apply (or ,debug-fn #'defaut-wrapper-print)
			  (list ,debug result))))
	(otherwise nil))
      result))

 (defvar *cond-defs* nil)
 (defvar *event-defs* nil)
 (defvar *action-defs* nil)

 (defmacro defcond (name body)
   (progn (if (assoc name *cond-defs*)
	      (status-report "~%Redefining ~a." name))
	  `(setq *cond-defs* (cons '(,name . ,body) *cond-defs*))))
 (defmacro defcondition (name body)
   (progn (if (assoc name *cond-defs*)
	      (status-report "~%Redefining ~a." name))
	  `(setq *cond-defs* (cons '(,name . ,body) *cond-defs*))))
 (defmacro defevent (name body)
   (progn (if (assoc name *event-defs*)
	      (status-report "~%Redefining ~a." name))
	  `(setq *event-defs* (cons '(,name . ,body) *event-defs*))))
 (defmacro defact (name body)
   (progn (if (assoc name *action-defs*)
	      (status-report "~%Redefining ~a." name))
	  `(setq *action-defs* (cons '(,name . ,body) *action-defs*))))
 (defmacro defaction (name body)
   (progn (if (assoc name *action-defs*)
	      (status-report "~%Redefining ~a." name))
	  `(setq *action-defs* (cons '(,name . ,body) *action-defs*))))

 (defun cboundp (x) (assoc x *cond-defs*))
 (defun eboundp (x) (assoc x *event-defs*))
 (defun aboundp (x) (assoc x *action-defs*))

 ;; Robot support

 (defmacro status-report (&rest l)
   (list 'if '*report-mode*
	 (append '(format *terminal-io*) l)
	 nil))

 )

(defun cleardefs ()
  (setq *cond-defs* nil)
  (setq *event-defs* nil)
  (setq *action-defs* nil))
     
(defun comment (string) (status-report string))

(defun lookup (x l)
  (let ((pair (assoc x l)))
    (if pair
	(cdr pair)
	(error "LOOKUP: unbound variable ~a." x))))

(defun function-def (x)
  (function-def1 x (list *cond-defs* *event-defs* *action-defs*)))

(defun function-def1 (x l)
  (if l
      (let ((body (lookup x (car l))))
	(if body body
	    (function-def1 x (cdr l))))
      (status-report "~%Undefined state function ~a." x)))

(defvar vmap nil)

(defun find-duplicate (l)
  (cond ((atom l) nil)
	((member (car l) (cdr l))
	 (car l))
	(t (find-duplicate (cdr l)))))

(defun no-duplicatesp (l)
  (cond ((atom l) t)
	(t (and (not (member (car l) (cdr l)))
		(no-duplicatesp (cdr l))))))

(defun setp (l) (no-duplicatesp l))

(defun bag-to-set (l)
  (cond ((atom l) nil)
	((member (car l) (cdr l)) (bag-to-set (cdr l)))
	(t (cons (car l) (bag-to-set (cdr l))))))

(defun bags-to-set (l)
  (if (atom l)
      nil
      (union (bag-to-set (car l)) (bags-to-set (cdr l)))))

(defun domain (l) (mapcar #'car l))

(defun range (l) (mapcar #'cdr l))

(defun non-null-symbolp (x)
  (and (symbolp x) (not (null x))))

(defun naturalp (x) 
  (and (integerp x)
       (<= 0 x)))

(defun non-null-symbol-setp (l)
  (and (listp l)
       (every #'non-null-symbolp l)
       (no-duplicatesp l)))

(defun unionx (l)
  (cond ((atom l) nil)
	(t (union (car l) (unionx (cdr l))))))

(defun intersectionx (l)
  (cond ((atom l) nil)
	((atom (cdr l)) (car l))
	(t (intersection (car l) (intersectionx (cdr l))))))

(defun unionx-equal (l)
  (cond ((atom l) nil)
	(t (union (car l) (unionx-equal (cdr l)) :test #'equal))))

(defun bind (x y a)
  (cond ((atom a)
	 (list (cons x y)))
	((equal (caar a) x)
	 (cons (cons x y) (cdr a)))
	(t (cons (car a)
		 (bind x y (cdr a))))))

(defun bind-all (assigns alist)
  (if (atom assigns)
      alist
      (bind-all (cdr assigns) (bind (caar assigns) (cdar assigns) alist))))

(defun binding (x a)
  (cdr (assoc x a :test #'equal)))

(defun bound? (x a)
  (assoc x a :test #'equal))
  
(defun all-bound? (l a)
  (every (f-l (x) (bound? x a)) l))

(defun put-nth (n v l)
  (cond ((atom l) l)
	((zerop n) (cons v (cdr l)))
	(t (put-nth (1- n) v (cdr l)))))

(defun list-product (l)
  (cond ((atom l) (list nil))
	(t (iterate for x in (car l)
	     append (iterate for y in (list-product (cdr l))
		       collect (cons x y))))))

;; UPDATE-FIELDS is used to define an update macro for a defstruct.
;; name is a defstruct name
;; obj is an object of the structure type
;; fieldnames is a list of fieldnames of the structure
;; updates is a list of alternating keywords and values, where
;;   the keywords are fieldnames
;; UPDATE-FIELDS returns an arglist suitable for the defstruct make function.
;
;; Example of use.
;
;; (defstruct foo a b)
;; (defmacro update-foo (expr &rest rest) 
;;    `(make-foo ,@(update-fields 'foo expr '(a b) rest)))
;
;; Then
;; (update-foo (make-foo) :b 4) = #S(FOO A NIL B 3)

(defun update-fields (name obj fieldnames updates)
  (iterate for fieldname in fieldnames
    append (let* ((keyword (intern (format nil "~a" fieldname) 'keyword))
 		  (l (member keyword updates)))
 	     (if (consp l)
 		 `(,keyword ,(cadr l))
	       `(,keyword (,(intern (format nil "~a-~a" name fieldname)) ,obj)))
 	     )))

(defun merge-alists (a b &optional chk-overwrite)
  (cond ((atom a) b)
	(chk-overwrite
	 (if (and (bound? (caar a) b)
		  (not (equal (binding (caar a) b) (cdar a))))
	     (error "MERGE-ALISTS. Overwriting ~a." (caar a))
	   (bind (caar a) (cdar a)
		 (merge-alists (cdr a) b chk-overwrite))))
	(t (bind (caar a) (cdar a)
		 (merge-alists (cdr a) b  chk-overwrite)))))

;; 
(defun local-var-reader () (list 'lookup (list 'quote (read)) 'lmap))
(defvar lmap nil)
(defun global-var-reader () (list 'lookup (list 'quote (read)) 'vmap))
(set-macro-character #\$ #'local-var-reader)
(set-macro-character #\! #'global-var-reader)
(defun localp (x)
  (and (consp x)
       (eq (car x) 'lookup)
       (and (consp (cddr x)) (eq (caddr x) 'lmap))
       (symbolp (cadr x))))
(defun var (x)
  (if (and (consp x)
	   (eq (car x) 'lookup)
	   (consp (cdr x)))
      (cadr x)
      nil))
(defun globalp (x)
  (and (consp x)
       (eq (car x) 'lookup)
       (and (consp (cddr x)) (eq (caddr x) 'vmap))
       (symbolp (cadr x))))
