
;; ------------------------------------------------------------
;; William R. Bevier and Michael K. Smith.
;; Copyright 1994, Computational Logic, Inc. All Rights Reserved
;; ------------------------------------------------------------

(require "nodes")
(require "arcs")
(require "system")
(provide "io")

;; ------------------------------------------------------------
;; INITIAL-SYSTEM: Main interface. Build a SYSTEM from the initial inputs
;;    node tuples
;;    arc tuples
;;    rootname
;;    initial configuration
;; ------------------------------------------------------------

(defun initial-system (ntuples atuples rootname 
			       &key 
			       initial-configuration
			       initial-alist
			       prim-events)
  (let ((nodes (make-nodes ntuples rootname)))
    (let ((arcs (make-arcs atuples nodes)))
      (make-system :nodes nodes
		   :root rootname
		   :arcs arcs
		   :configuration (full-configuration
				   (or initial-configuration (list rootname))
				   nodes)
		   :alist initial-alist
		   :prim-events prim-events))))

(defun system-eventp (expr system states)
  (eventp expr (system-prim-events system) states (system-prim-vars system)))

(defun system-actionp (expr system states)
  (actionp expr (system-prim-events system) states (system-prim-vars system)))
					      

;; ------------------------------------------------------------
;; Entering the node hierarchy.
;; An input node set is a list with elements of the form
;
;;     (nodename [type] [children] [enter] [exit] [initial])
;
;; where bracketed fields are optional.
;
;; When type and children are missing, the tuple is interpreted
;; as a basic node.
;
;; Examples:
;
;;     '(s and (a b c))
;
;;     '(c)
;
;; From this, MAKE-NODES constructs a list of node
;; structures, which stores frequently used information for 
;; each node.
;; ------------------------------------------------------------

;; MAKE-NODES is the main interface function for constructing the 
;; internal form of nodes from an input list of tuples.

(defun make-nodes (tuples rootname)
  (cond ((no-duplicatesp (mapcar #'car tuples))
	 (let ((nodes (make-nodes1 tuples)))
	   (cond ((node-listp nodes)
		  (let ((nodes (compute-static-node-information nodes)))
		    (if (full-node-listp nodes rootname)
			nodes
			(error "MAKE-NODES: Invalid full set of nodes."))))
		 (t (error "MAKE-NODES: Invalid set of tuples.")))))
	(t (error "MAKE-NODES: Duplicate node name ~a."
		  (find-duplicate (mapcar #'car tuples))))))

;; MAKE-NODES1 builds a basic node list
;; from a list of node tuples as described above. 
;; Only the type and children of a node are recorded.

(defun make-nodes1 (tuples)
  (iterate for tuple in tuples
    collect (make-node :name (nth 0 tuple)
		       :type (if (nth 1 tuple) (nth 1 tuple) 'basic)
		       :children (nth 2 tuple)
		       ;; :idle -1
		       ;; :busy -1
		       :enter (nth 3 tuple)
		       :exit (nth 4 tuple)
		       :initial (nth 5 tuple))))


;; ------------------------------------------------------------
;; Entering Arcs
;
;; An arc set is a list of lists of the form
;
;;         (arcname from tos [trigger] [action]).
;
;; From a list of such tuples, MAKE-ARCS constructs a list
;; of arc structures.
;; ------------------------------------------------------------

(defun make-arcs (tuples nodes)
  (cond ((no-duplicatesp (mapcar #'car tuples))
	 (let ((arcs (make-arcs1 tuples)))
	   (cond ((arc-listp arcs nodes)
		  (let ((arcs (compute-static-arc-information arcs nodes)))
		    (if (full-arc-listp arcs nodes)
			arcs
			(error "MAKE-ARCS: Invalid set of full arcs."))))
		 (t (error "MAKE-ARCS: Invalid set of tuples.")))))
	(t (error "MAKE-ARCS: Duplicate arc name ~a."
		  (find-duplicate (mapcar #'car tuples))))))

(defun make-arcs1 (tuples)
  (iterate for x in tuples
	   collect (make-arc :name (nth 0 x)
			     :from (nth 1 x)
			     :to (nth 2 x)
			     :trigger (nth 3 x)
			     :action (nth 4 x))))

(defun compute-static-arc-information (arcs nodes)
  (iterate for a in arcs
    collect (update-arc a :scope (scope-of-arc a nodes))))

(defun delete-arcs (names arcs)
  (iterate for a in arcs
    when (not (member (arc-name a) names))
    collect a))

;; EXITING-ARCS
;; Find the names of all arcs exiting a node named NM.

(defun exiting-arcs (nm arcs)
  (iterate for a in arcs
    when (equal nm (arc-from a))
    collect a))


