; ------------------------------------------------------------
;; William R. Bevier and Michael K. Smith.
;; Copyright 1994, 1995, 1996 Computational Logic, Inc. All Rights Reserved
;; ------------------------------------------------------------

;; From The Semantics of Statecharts
;; Section: States and Configurations

(require "general")
(provide "nodes")

;; ------------------------------------------------------------
;; Changes for Version2
;
;; We introduce Node as a generalization of states. Nodes include
;; states, as well as various kinds of connectors, and terminators.
;; Nodes are the abstract sources and destinations of arcs.
;
;; In the internal form of a SYSTEM, only states are the 
;; sources and targets of arcs.
;; ------------------------------------------------------------

;; Node are named entities, each represented by a NODE structure.

(defconstant *node-types* '(BASIC AND OR))

(eval-when
 (compile load eval)

 (defstruct node
   name					; a symbol, the state name (given)
   type					; a symbol, the state type (given)
					; `state' nodes have type BASIC, AND or OR.
					; BASIC nodes have no children.
   (level 0)				; level of NODE in the hierarchy
   parent				; name of NODE parent
   children				; names of NODE's children
   initial				; name of initial active child if NODE is an OR. 

   (idle -1)				; Time of last entry. Reset to -1 on exit.
					; A busy loop reset this each cycle.
   (busy -1)				; -1, if not active. 0, if just entered.
					; Else the number of busy loops.
   enter
   exit

   ancestors
   descendents)

 (defmacro update-node (x &rest rest)
   `(make-node ,@(update-fields 'node
				x
				'(name type level parent children
				       initial idle busy enter exit ancestors descendents)
				rest)))
 )

;; An `initial node' is a child of an OR that is activated 
;; simulateously with its parent.

;; Forall nodes, 
;;   NODEP must be true before (COMPUTE-STATIC-NODE-INFORMATION nodes), 
;;   FULL-NODEP after.

(defun nodep (x)
  (and (static-check (node-p x)
		    "~%NODEP: ~a not a node struct." x)
       (static-check (non-null-symbolp (node-name x))
		    "~%NODEP: Node name ~a should be a non-nil symbol." (node-name x))
       (static-check (member (node-type x) *node-types*)
		    "~%NODEP: Node ~a does not have a level type: ~a." (node-name x) (node-type x))
       (static-check (naturalp (node-level x))
		    "~%NODEP: Node ~a's level is not a natural number." (node-name x))
       (static-check (or (not (node-initial x)) (symbolp (node-initial x)))
		    "~%NODEP: Node ~a's initial child is not a non-nil symbol" (node-name x))
       (static-check (or (not (node-initial x)) (member (node-initial x) (node-children x)))
		    "~%NODEP: Node ~a's initial node is not a child" (node-name x))
       (static-check (or (not (node-initial x)) (equal (node-type x) 'OR))
		    "~%NODEP: Node ~a has an initial node, but is not and OR." (node-name x))
       (static-check (non-null-symbol-setp (node-children x))
		    "~%NODEP: Node ~a's children is not a list of non-nil symbols" (node-name x))
       (static-check (or (member (node-type x) '(and or))
			(null (node-children x)))
		    "~%NODEP: Node ~a has children, but is not and AND or OR node." (node-name x))))

(defun full-nodep (node nodes rootname)
  " rootname: nodename "
  (let ((nm (node-name node))
	(names (node-names nodes)))
    (and (nodep node)
	 (static-check (or (eq nm rootname) (non-null-symbolp (node-parent node)))
		      "~%FULL-NODEP: Node ~a's parent is not a non-nil symbol" (node-name node))
	 (static-check (every (f-l (x) (member x names)) (node-children node))
		      "~%FULL-NODEP: Node ~a has a child that is not a node name." nm)
	 (static-check (eql (length (all-parents nm nodes))
			   (if (eq nm rootname) 0 1))
		      "~%FULL-NODEP: Node ~a has too many parents: ~a."
		      nm (all-parents nm nodes)))))

(defun node-listp (nodes)
  " Nodes   : list of nodes "
  (iterate for nd in nodes
	   always (nodep nd)))

(defun full-node-listp (nodes rootname)
  " nodes   : list of nodes 
 rootname: nodename "
  (if (static-check (non-null-symbolp rootname)
		   "~%FULL-NODE-LISTP: The root name, ~a, is not a non NIL symbolp." rootname)
      (iterate for nd in nodes
	       always (full-nodep nd nodes rootname))
      t))

;; The following functions compute various information
;; about a named nodes from a list of node structs.

;; These functions rely only on the type and children field. 
;; They are used to compute information that is later stored
;; in a node struct for fast retrieval.

(defun all-parents (nm nodes)
  " nm: nodename 
 nodes: list of nodes "
  (iterate for nd in nodes
    when (member nm (node-children nd))
    collect (node-name nd)))

(defun node-names (nodes)
  (iterate for nd in nodes
    collect (node-name nd)))

(defun parent (nm nodes)
  (iterate for nd in nodes
    when (member nm (node-children nd))
    return (node-name nd)))

(defun ancestors (nm nodes)
  (let ((p (parent nm nodes)))
    (cond ((eq p nil) nil)
	  (t (cons p (ancestors p nodes))))))

(defun find-node (nm nodes)
  (iterate for nd in nodes
    when (eq nm (node-name nd))
    return nd))

(defun find-nodes (nms nodes)
  (iterate for nd in nodes
	   when (member (node-name nd) nms)
	   collect nd))

(defun children (nm nodes)
  (node-children (find-node nm nodes)))

(defun descendents (nm nodes)
  (let ((c (children nm nodes)))
    (cond ((null c) nil)
	  (t (iterate for x in c
	       append (cons x (descendents x nodes)))))))

(defun configurationp (names nodes)
  (iterate for nm in names
    always (let* ((nd (find-node nm nodes))
		  (type (node-type nd))
		  (children (node-children nd)))
	     (case type
	       (or (eql (length (intersection children names)) 1))
	       (and (eql (length (intersection children names))
			 (length children)))
	       (basic t)
	       (t (error "CONFIGURATIONP: Bad type ~a ~a" nm type))))))
			      
(defun full-configuration (names nodes)
  (unionx (cons names
		(iterate for nm in names
		  collect (ancestors nm nodes)))))

(defun basic-configurationp (names nodes)
  (and (iterate for nm in names
	 always (eq (node-type (find-node nm nodes)) 'basic))
       (configurationp (full-configuration names nodes) nodes)))

(defun level-of-node (nm nodes)
  (length (ancestors nm nodes)))

(defun deepest-node (names nodes)
  " The deepest node in NAMES. "
  (cond ((atom names) nil)
	((atom (cdr names)) (car names))
	(t (let ((nm (deepest-node (cdr names) nodes)))
	     (if (< (nlevel nm nodes)
		    (nlevel (car names) nodes))
		 (car names)
	       nm)))))

(defun deepest-containing-node (names nodes)
  " The name of the deepest node in NODES that 
 properly contains all the nodes in NAMES."
  (let ((sets (iterate for nm in names  
		collect (nancestors nm nodes))))
    (deepest-node (intersectionx sets) nodes)))

(defun deepest-containing-OR-node (nm nodes)
  (if (or (null nm)
	  (eq (ntype nm nodes) 'or))
      nm
    (deepest-containing-OR-node (nparent nm nodes) nodes)))

;; COMPUTE-STATIC-NODE-INFORMATION updates a node list to record 
;; information about a node that is statically true, e.g. its set of ancestors.

(defun compute-static-node-information (nodes)
  (iterate for nd in nodes
    collect (let* ((nm (node-name nd))
		   (level (level-of-node nm nodes)))
	      (update-node nd
			   :level level
			   :parent (parent nm nodes)
			   :ancestors (ancestors nm nodes)
			   :descendents (descendents nm nodes)))))

;; ------------------------------------------------------------
;; Accessor and update functions for lists of nodes
;; ------------------------------------------------------------

;; Accessor functions. Get the indicated field for the node
;; with name N.

(defun ntype (nm nodes)
  (node-type (find-node nm nodes)))

(defun or-nodep (node)
  (eq (node-type node) 'or))

(defun and-nodep (node)
  (eq (node-type node) 'and))

(defun basic-nodep (node)
  (eq (node-type node) 'basic))

(defun nlevel (nm nodes)
  (node-level (find-node nm nodes)))

(defun nparent (nm nodes)
  (node-parent (find-node nm nodes)))

(defun nchildren (nm nodes)
  (node-children (find-node nm nodes)))

(defun ninitial (nm nodes)
  (node-initial (find-node nm nodes)))

(defun nancestors (nm nodes)
  (node-ancestors (find-node nm nodes)))

(defun ndescendents (nm nodes)
  (node-descendents (find-node nm nodes)))

(defun nidle (nm nodes)
  (node-idle (find-node nm nodes)))

(defun nbusy (nm nodes)
  (node-busy (find-node nm nodes)))

(defun nenter (nm nodes)
  (node-enter (find-node nm nodes)))

(defun nexit (nm nodes)
  (node-exit (find-node nm nodes)))

(defun nancestors* (l nodes)
  (iterate for x in l
    append (nancestors x nodes)))

(defun ndescendents* (l nodes)
  (iterate for x in l
    append (ndescendents x nodes)))



